import argparse
import os
import sys
import multiprocessing

from .loader import load_class
from .task import AkasakaTask
from .torch_task import AkasakaTorchTask


def main():
    parser = argparse.ArgumentParser(description='Dynamically load a Python class from a module.', add_help=False)
    parser.add_argument('module_path', type=str, help='The module path in the format "module.submodule.ClassName"')
    parser.add_argument('--num_process', type=int, default=multiprocessing.cpu_count(), help='Number of processes to use, for normal task (default: number of CPU cores)')
    parser.add_argument('--chunksize', type=int, default=1, help='Number of tasks to be sent to a worker process at a time, for normal task (default: 1)')
    parser.add_argument('--devices', type=int, nargs='+', default=[], help='The devices to run the task on, for torch task (default: [])')
    parser.add_argument('-h', '--help', action='store_true', help='Show help message for task and exit')
    parser.add_argument('-H', '--hel-akasaka', action='help', help='Show Akasaka help message and exit')
    args, remaining_args = parser.parse_known_args()
    
    module_path = args.module_path
    
    if args.help:
        remaining_args += ["-h"]

    # Add the current working directory to the sys.path so that the module can be found
    sys.path.append(os.getcwd())
    # Load the class from the module
    loaded_class = load_class(module_path)
    print(f"Successfully loaded class '{loaded_class.__name__}' from module '{args.module_path}'.")

    # Initialize the task with the remaining command line arguments
    akasaka_task = loaded_class(args=remaining_args)

    if issubclass(loaded_class, AkasakaTask):
        from .executor import akasaka_execute
        num_process = args.num_process
        chunksize = args.chunksize
        akasaka_execute(akasaka_task, num_process, chunksize)
        return

    if issubclass(loaded_class, AkasakaTorchTask):
        devices = args.devices
        if len(devices) == 1:
            from .torch_single_executor import akasaka_torch_single_execute
            akasaka_torch_single_execute(akasaka_task, devices[0])
        else:
            from .torch_executor import akasaka_torch_execute
            akasaka_torch_execute(akasaka_task, module_path, remaining_args, devices)
        return

    raise ValueError("The loaded class should be a subclass of 'AkasakaTask' or 'AkasakaTorchTask'.")



import multiprocessing
from tqdm import tqdm


akasaka_task = None


def execute(args):
    global akasaka_task

    if akasaka_task.is_executed(**args):
        return
    akasaka_task(**args)


def akasaka_execute(task, num_process, chunksize):
    global akasaka_task
    akasaka_task = task
    # Generate a list of tasks to be executed
    tasks = akasaka_task.generate_tasks()
    tasks = list(tasks)
    print(f"Generated {len(tasks)} tasks to be executed.")

    pool = multiprocessing.Pool(processes=num_process)
    _ = list(tqdm(pool.imap_unordered(execute, tasks, chunksize=chunksize), total=len(tasks)))

    pool.close()
    pool.join()



class AkasakaTorchTask():
    def __init__(self, args) -> None:
        """
        Initialize the task with command line arguments
        args can be passed to ArgumentParser.parse_args() to parse remaining command line arguments.
        """
        pass

    def load_model(self, device) -> None:
        """Load the model on a device."""
        raise NotImplementedError("This method should be implemented in a subclass.")

    def get_dataloader(self):
        """Return a DataLoader object."""
        raise NotImplementedError("This method should be implemented in a subclass.")

    def is_executed(self, batch) -> bool:
        """Check if the task has already been executed."""
        raise NotImplementedError("This method should be implemented in a subclass.")

    def __call__(self, device, batch) -> None:
        """Run the task. Also provide the device to run the task on."""
        raise NotImplementedError("This method should be implemented in a subclass.")



import os
import glob
from .helper import make_consecutive_pairs


def walk_files_of_dir(root: str, ext=None):
    """Walk through all files in a directory

    Args:
        root (str): root directory to start from
        ext (list[str], optional): extension names to filter. Defaults to None.

    Returns:
        dict: directory (relative path) -> list of files
    """
    files = {}

    # Walk through the directory tree
    for dirpath, _, filenames in os.walk(root):
        # Add directory path to the dictionary
        relative_path = os.path.relpath(dirpath, root)
        filenames = sorted(filenames)
        # Filter files by extension
        if ext:
            filenames = [f for f in filenames if str(f).lower().endswith(tuple(ext))]
        files[relative_path] = filenames

    return files


def walk_pairs_of_dir(root: str, ext=None):
    """Walk through all pairs of files in a directory

    Args:
        root (str): root directory to start from
        ext (list[str], optional): extension names to filter. Defaults to None.

    Returns:
        dict: directory (relative path) -> list of pairs of files
    """
    files = walk_files_of_dir(root, ext)

    # Make consecutive pairs of files
    pairs = {relative_path: make_consecutive_pairs(files) for relative_path, files in files.items()}

    return pairs



import importlib.util


def load_class(module_path):
    try:
        module_name, class_name = module_path.rsplit('.', 1)
        spec = importlib.util.find_spec(module_name)
        if spec is None:
            raise ImportError(f"Module '{module_name}' not found.")
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        if not hasattr(module, class_name):
            raise AttributeError(f"Class '{class_name}' not found in module '{module_name}'.")
        return getattr(module, class_name)
    except ValueError:
        raise ValueError("Invalid module path. It should be in the format 'module.submodule.ClassName'.")



import torch
from tqdm import tqdm


def akasaka_torch_single_execute(task, device):
    if not torch.cuda.is_available():
        raise ValueError("CUDA is not available.")

    task.load_model(device)
    dataloader = task.get_dataloader()

    for batch in tqdm(dataloader):
        if task.is_executed(batch):
            continue
        task(device=device, batch=batch)



def make_consecutive_pairs(lst):
    """Make consecutive pairs from a list

    Args:
        lst (list): list to make pairs from

    Returns:
        list: list of consecutive pairs
    """
    return list(zip(lst[:-1], lst[1:]))



import torch
from tqdm import tqdm
import torch.multiprocessing as mp
from .loader import load_class


# these two variables only initialize
# after the `load` function is called
# i.e. in the new process
akasaka_task = None
device = None


def load(module_path, args, cuda_devices):
    # we need to reload the module in the new process
    # because pytorch only supports spawn, which will start with a fresh python interpreter
    loaded_class = load_class(module_path)
    task = loaded_class(args=args)
    # get the process number -> device index
    d_index = mp.current_process()._identity[0] - 1
    # load the model to the device of this process
    task.load_model(cuda_devices[d_index])
    print(f"Process {d_index} finished loading model.")

    # make it global of this process
    global akasaka_task
    akasaka_task = task
    global device
    device = cuda_devices[d_index]
    print(f"Process {d_index} finished globaling.")


def execute(batch):
    # access the global task and device (pre-process)
    global akasaka_task
    global device

    if akasaka_task.is_executed(batch):
        return

    akasaka_task(device=device, batch=batch)


def akasaka_torch_execute(task, module_path, args, devices):
    if not torch.cuda.is_available():
        raise ValueError("CUDA is not available.")

    # create torch devices from the device indices
    cuda_devices = []
    for device in devices:
        cuda_devices.append(torch.device(f"cuda:{device}"))

    # WE ONLY USE TASK TO GET DATALOADER -> distribute the data to each process later
    dataloader = task.get_dataloader()

    # WE WILL RELOAD TASK MODULE IN EACH PROCESS
    mp.set_start_method('spawn')
    pool = mp.get_context('spawn').Pool(processes=len(devices), initializer=load, initargs=(module_path, args, cuda_devices))
    _ = list(tqdm(pool.imap_unordered(execute, dataloader), total=len(dataloader)))

    pool.close()
    pool.join()



class AkasakaTask():
    def __init__(self, args) -> None:
        """
        Initialize the task with command line arguments
        args can be passed to ArgumentParser.parse_args() to parse remaining command line arguments.
        """
        pass

    def generate_tasks(self) -> list:
        """Generate a list of tasks to be executed."""
        raise NotImplementedError("This method should be implemented in a subclass.")

    def is_executed(self, **kwargs) -> bool:
        """Check if the task has already been executed."""
        raise NotImplementedError("This method should be implemented in a subclass.")

    def __call__(self, **kwargs) -> None:
        """Run the task."""
        raise NotImplementedError("This method should be implemented in a subclass.")



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("version", "r", encoding="utf-8") as fh:
    version = fh.read().strip()

def make_email(name, domain):
    return f"{name}@{domain}"


setup(
    name="akasaka",
    version=version,
    author="Haoyun Qin",
    author_email=make_email("qhy.cis", "gmail.com"),
    description="Dynamic mutiprocess preprocessing task loader and dispatcher",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JeffersonQin/akasaka",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'akasaka = akasaka.main:main'
        ]
    },
    install_requires=[
        'tqdm'
    ]
)
